import itertools
from .apiUtils import safeMapResToJson, concurrentExecRequests
from .itisTools import commonToScientific, \
                       scientificToCommon, \
                       tsnMatching, \
                       synonymMatching, \
                       downstreamHierarchy

class Itis:
    def __init__(self):
        pass 

    @staticmethod
    def commonToScientificName(names):
        urls = list(map(commonToScientific.constructUrls,names))
        results = concurrentExecRequests(urls)
        results = safeMapResToJson(results)
        results = list(zip(names,results))
        results = list(map(commonToScientific.processIndividualResult,results))
        return results

    @staticmethod
    def scientificToCommonName(names):
        urls = list(map(scientificToCommon.constructUrls,names))
        results = concurrentExecRequests(urls)
        results = safeMapResToJson(results)
        results = list(zip(names,results))
        results = list(map(scientificToCommon.processIndividualResult,results))
        return results
    
    @staticmethod
    def retrieveTSNs(names):
        urls = list(map(tsnMatching.constructUrls,names))
        results = concurrentExecRequests(urls)
        results = safeMapResToJson(results)
        results = list(zip(names,results))
        results = list(map(tsnMatching.processIndividualResult,results))
        return results
    
    @staticmethod
    def getSynonymsFromTSNs(tsns):
        urls = list(map(synonymMatching.constructUrls,tsns))
        results = concurrentExecRequests(urls)
        results = safeMapResToJson(results)
        results = list(zip(tsns,results))
        results = list(map(synonymMatching.processIndividualResult,results))
        return results
    
    @staticmethod
    def downstreamToNextRank(tsns):
        urls = list(map(downstreamHierarchy.constructUrls,tsns))
        results = concurrentExecRequests(urls)
        results = safeMapResToJson(results)
        results = list(zip(tsns,results))
        return list(map(downstreamHierarchy.processIndividualResult,results))
        




from ..externalAPIs import Itis
import itertools

def commonToScientific(names):
    try:
        return Itis.commonToScientificName(names)
    except Exception as e:
        print("Caught fatal exception in common name resolver")
        print(str(e))
        return list(map(lambda x: (x[0],[]),names))





from .services import createIndex, \
                      crushIndexInput, \
                      writeApiResultsToIndex, \
                      ensureUniqueInIndex, \
                      findDownstreamChildren, \
                      tagInputWithTaxaLevel, \
                      findDownstreamChildren, \
                      taxa_ordering 
                      
from collections import defaultdict

def children(names):
    index = createIndex(names)
    taggedInput = tagInputWithTaxaLevel([x[0] for x in crushIndexInput(index)])
    groupedByLevel = defaultdict(list)
    writeApiResultsToIndex(index,taggedInput) 
    nameAndRankTuple = crushIndexInput(index)
    for cleanedName,(rank,value) in nameAndRankTuple:
        groupedByLevel[rank].append((cleanedName,(rank,value)))

    for rank in groupedByLevel:
        groupedByLevel[rank] = list(zip(first(groupedByLevel[rank]),findDownstreamChildren(second(groupedByLevel[rank]),nextLowestRank(rank))))
    
    results = []
    for rank in groupedByLevel:
        for cleanedName,(correctedName,res) in groupedByLevel[rank]:
            results.append((cleanedName,res))
    writeApiResultsToIndex(index,results)
    return index

def second(nameRankTuples):
    return list(map(lambda x: x[1], nameRankTuples))

def first(nameRankTuples):
    return list(map(lambda x: x[0], nameRankTuples))

def nextLowestRank(rank):
    previous = "species"
    for item in taxa_ordering:
        if item == rank:
            return previous
        previous = item
    return "species"
    


def constructTaxaProfilingUrl(name):
    return 'https://api.gbif.org/v1/species?name='+name

def processTaxaProfilingIndividualResult(nameResTuple):
    name,response = nameResTuple
    return (name,response['result'])


from ..externalAPIs import Gnr
import itertools
from .constants import taxa_ordering
from .services import writeApiResultsToIndex

def profile_taxonomy(names):
    try:
        return Gnr.scientificNameToFullTaxonomy(names)
    except Exception as e:
        print("Caught fatal exception in taxonomic indexer")
        print(str(e))
        return {}    




def constructUrls(nameAndTaxaTuple):
    rank,value = nameAndTaxaTuple
    return 'https://api.gbif.org/v1/species/search?q='+value+'&rank=species'

def processIndividualResult(identifiedResponse):
    inTuple, response = identifiedResponse
    rank, value = inTuple
    if 'results' in response:
        potentialSpecies = list(map(lambda x: x.get('species',''),response['results']))
        potentialSpecies = list(filter(lambda x: x != '',potentialSpecies))
        if rank == "genus": potentialSpecies = list(filter(lambda x: x.split(" ")[0].lower() == value.lower(),potentialSpecies))
        return (rank,value,list(set(potentialSpecies)))
    return (rank,value,[])


import itertools 

def constructUrls(name):
    return "https://itis.gov/ITISWebService/jsonservice/getITISTermsFromCommonName?srchKey="+name

def processIndividualResult(nameAndResTuple):
    try:
        name,res = nameAndResTuple
        if 'itisTerms' not in res: return (name,[])
        if res['itisTerms'] is None: return (name,[])
        if res['itisTerms'][0] is None: return (name,[])
        nameTranslation = res['itisTerms']
        scientificNames = processTranslationToScientificName(name,nameTranslation)
        scientificNames = expandIfHigherLevelTaxaName(scientificNames)
        return (name,scientificNames)
    except Exception as e:
        return (name,[])

def processTranslationToScientificName(name,nameTranslation):
    sNames = list(filter(containsUsableData,nameTranslation))
    sNames = list(filter(lambda x: containsValidCommonName(name,x['commonNames']), sNames))
    return list(map(lambda x: x['scientificName'],sNames))

def expandIfHigherLevelTaxaName(scientificNames):
    return list(filter(lambda x: len(x.strip().split(" ")) > 1,scientificNames))

def containsUsableData(individualCommonNameResult):
    return \
        'scientificName' in individualCommonNameResult and \
        'commonNames' in individualCommonNameResult and \
        individualCommonNameResult['commonNames'][0] is not None and \
        individualCommonNameResult['scientificName'] is not None

def containsValidCommonName(name,listOfCommonNames):
    if len(name.strip().split(" ")) > 1: return True
    return any(list(map(lambda x: checkIfCommonNameResultIsValid(name,x),listOfCommonNames)))

def checkIfCommonNameResultIsValid(name,result):
    result = result.lower().split(" ")
    stringPotential1 = name 
    stringPotential2 = name+'s' #plurals
    return stringPotential1 in result or stringPotential2 in result


def constructUrls(req):
    return f'http://resolver.globalnames.org/name_resolvers.json?names={req}'

def processIndividualResult(nameResTuple):
    name,response = nameResTuple
    topLevelMapping = (name,response \
                                .get('data',[{}]) \
                                [0] \
                                .get('results',[])) 
    return topLevelMapping

def parseSingleTaxonomyFromAPI(nameResTuple):
    name,taxonomicAPIres = nameResTuple
    dataFromMultipleSources = list(map(extractTaxaData,taxonomicAPIres))
    return (name,dataFromMultipleSources)
    
def extractTaxaData(singleTaxaSource):
    mappingDict = {}
    try: mappingDict = dict(zip(singleTaxaSource['classification_path_ranks'].lower().split("|"),\
                                singleTaxaSource['classification_path'].lower().split("|")))
    except: pass
    return mappingDict



from .apiUtils import safeMapResToJson,concurrentExecRequests
from .gbifTools import taxaToSpecies,taxaProfiling,diveDownTaxaTree

class Gbif:
    def __init__(self):
        pass 

    @staticmethod
    def higherTaxaToSpecies(originalNamesWithTaxa):
        urls = list(map(taxaToSpecies.constructUrls,originalNamesWithTaxa))
        results = concurrentExecRequests(urls)
        results = safeMapResToJson(results)
        results = list(zip(originalNamesWithTaxa,results))
        results = list(map(taxaToSpecies.processIndividualResult,results))
        return results

    @staticmethod
    def scientificNameToFullTaxonomy(names):
        capNames = list(map(lambda x: x.capitalize(),names))
        urls = list(map(taxaProfiling.constructTaxaProfilingUrl,capNames))
        results = concurrentExecRequests(urls)
        results = safeMapResToJson(results)
        results = list(zip(names,results))
        results = list(map(taxaProfiling.processTaxaProfilingIndividualResult,results))
        return results
    
    @staticmethod
    def downstreamToRank(nameRankTuples,rank):
        capNames = list(map(lambda x: (x[0].capitalize(),x[1]),nameRankTuples))
        urls = list(map(lambda x: diveDownTaxaTree.constructUrls(x,rank),capNames))
        results = concurrentExecRequests(urls)
        results = safeMapResToJson(results)
        results = list(zip(nameRankTuples,results))
        results = list(map(lambda x: diveDownTaxaTree.processIndividualResult(x,rank),results))
        return results



def constructUrls(name):
    return "https://www.itis.gov/ITISWebService/jsonservice/searchByScientificName?srchKey="+name

def processIndividualResult(nameAndResTuple):
    name,res = nameAndResTuple
    errorCode = (name,'-1')
    if 'scientificNames' not in res: return errorCode
    if res['scientificNames'] is None: return errorCode
    if res['scientificNames'][0] is None: return errorCode
    if len(res['scientificNames']) == 0: return errorCode

    for item in res['scientificNames']:
        if 'combinedName' not in item: continue
        if item['combinedName'] is None: continue
        if item['combinedName'].lower() == name.lower():
            if item['tsn'] is None: return errorCode 
            return (name,item['tsn'])
    
    return errorCode



from .constants import taxa_ordering
import itertools 
import operator
from collections import defaultdict
# "Back up" a claim of a taxonomic generalisation
# with at least ONE other source

def backed_up(nameResTuple):
    name,dataFromMultipleSources = nameResTuple
    if len(dataFromMultipleSources) < 2:
        return short_circuit(nameResTuple)
    taxonomicClasses = defaultdict(list)
    for item in taxa_ordering:
        for taxaMap in dataFromMultipleSources:
            if taxaMap.get(item,'').strip() != '': 
                taxonomicClasses[item].append(taxaMap[item].strip().lower())

    
    for item in taxa_ordering:
        if len(taxonomicClasses.get(item,'')) > 1: #percentage of data sources?
            return (name,(item,mostCommonInList(taxonomicClasses[item])))
    return short_circuit(nameResTuple)

def short_circuit(nameResTuple):
    name,dataFromMultipleSources = nameResTuple
    for item in taxa_ordering:
        for taxaMap in dataFromMultipleSources:
            if (item in taxaMap) and (taxaMap[item].strip() != ''): 
                return (name,(item,taxaMap[item].strip()))
    return (name,('',''))

def highest_count_wins_multi_cat(inputStruct):
    dataFromMultipleSources = []
    if len(inputStruct) == 2:
        name,dataFromMultipleSources = inputStruct
    else:
        dataFromMultipleSources = inputStruct
    grouped = {item: \
                mostCommonInList( \
                    list( \
                        filter( \
                            lambda groupVal: groupVal != '', \
                            map( \
                                lambda taxaDict: taxaDict.get(item,''), \
                                dataFromMultipleSources \
                            ) \
                        ) \
                    ) \
                )
                for item in taxa_ordering}
    
    if len(inputStruct) == 2:
        return (name,grouped)
    
    return grouped

def mostCommonInList(L,emptyOnDefault=''):
  if len(L) == 0: return emptyOnDefault
  SL = sorted((x, i) for i, x in enumerate(L))
  groups = itertools.groupby(SL, key=operator.itemgetter(0))
  def _auxfun(g):
    item, iterable = g
    count = 0
    min_index = len(L)
    for _, where in iterable:
      count += 1
      min_index = min(min_index, where)
    return count, -min_index
  return max(groups, key=_auxfun)[0]


from .services import highest_count_wins_multi_cat
from .classify import classify
from collections import Counter

def sanityCheckOutput(oIndex):
    index = indexToTuples(oIndex)
    enrichedResults = enrichSpeciesToFullTaxonomy(index)
    for k, (real, cleaned, lst) in enumerate(enrichedResults):
        groupingByFamily = summaryStatsPerCategory(lst).get("family", "")
        if len(groupingByFamily) < 2:
            enrichedResults[k] = (real, cleaned, oIndex[real][1])
        else:
            rankedGrouping = sorted(
                groupingByFamily.items(), key=lambda x: x[1], reverse=True
            )
            enrichedResults[k] = (
                real,
                cleaned,
                decideTranslationOnGroupStats(rankedGrouping, lst),
            )

    return dict(list(map(lambda x: [x[0], [*x[1:]]], enrichedResults)))


def enrichSpeciesToFullTaxonomy(index):
    return list(
        map(
            lambda x: [
                x[0],
                x[1],
                list(map(lambda y: y[1][1], list(classify(x[2]).items()))),
            ],
            index,
        )
    )

def decideTranslationOnGroupStats(rankedGrouping, taxaList):
    leading = rankedGrouping[0][1]
    total = sum([item[1] for item in rankedGrouping])
    if (leading / total) <= 0.75:
        out = rankedGrouping[1:]
        if not all([(val[1] <= 0.1 * total and leading >= 0.5 * total) for val in out]):
            return []

    return takeSpeciesMatchingGroupOnly(rankedGrouping[0][0], "family", taxaList)


def summaryStatsPerCategory(result):
    groups = {
        cat: list(map(lambda x: x.get(cat, ""), result)) for cat in ['family']
    }
    summaryStats = {k: grouping(v) for (k, v) in groups.items()}
    if "" in summaryStats:
        del summaryStats[""]
    return summaryStats


def grouping(values):
    baseResults = dict(Counter(values))
    if "" in baseResults:
        del baseResults[""]
    return baseResults


def indexToTuples(index):
    return list(map(lambda x: [x, *index[x]], list(index.keys())))


def takeSpeciesMatchingGroupOnly(group, level, data):
    species = []
    for item in data:
        if item.get(level, "") == group:
            if "species" in item:
                species.append(item["species"])

    return species



from .services import commonToScientific, \
                      tagInputWithTaxaLevel, \
                      createIndex, \
                      mapKnownTaxaLevelToSpecies, \
                      crushIndexInput, \
                      writeApiResultsToIndex, \
                      ensureUniqueInIndex, \
                      profile_taxonomy, \
                      highest_count_wins_multi_cat, \
                      findDownstreamChildren \

from .speciesValidation import sanityCheckOutput

def to_scientific(names,sanityCheck=False):
    index = createIndex(names)
    results = commonToScientific(names)
    writeApiResultsToIndex(index,results)
    if sanityCheck: 
        index = sanityCheckOutput(index)
    ensureUniqueInIndex(index)
    return index




def constructUrls(tsn):
    return 'https://www.itis.gov/ITISWebService/jsonservice/getHierarchyDownFromTSN?tsn='+tsn

def processIndividualResult(nameResTuple):
    try:
        name,result = nameResTuple
        if 'hierarchyList' not in result or len(result['hierarchyList']) == 0 or result['hierarchyList'][0] is None:
            return (name,[])
        
        return (name,list(map(lambda x: (x['rankName'].lower(),x['taxonName']),result['hierarchyList'])))
    except Exception as e:
        print("Caught fatal exception in downstream name API")
        print(str(e))
        return (name,[])




from .services import createIndex,\
                      crushIndexInput,\
                      writeApiResultsToIndex,\
                      ensureUniqueInIndex,\
                      findDownstreamChildren,\
                      tagInputWithTaxaLevel

def downstream(names,rank):
    index = createIndex(names)
    taggedInput = tagInputWithTaxaLevel([x[0] for x in crushIndexInput(index)])
    writeApiResultsToIndex(index,taggedInput) 
    nameAndRankTuple = crushIndexInput(index)
    results = list(findDownstreamChildren(nameAndRankTuple,rank))
    writeApiResultsToIndex(index,results)
    return index
    
def second(nameRankTuples):
    return list(map(lambda x: x[1], nameRankTuples))

def first(nameRankTuples):
    return list(map(lambda x: x[0], nameRankTuples))




def constructUrls(nameRankTuple,rank):
    givenRank,name = nameRankTuple
    return 'https://api.gbif.org/v1/species/search?q='+name+'&rank='+rank

def processIndividualResult(identifiedResponse,rank):
    nameRankTuple, response = identifiedResponse
    givenRank,name = nameRankTuple
    if 'results' in response:
        potentialSpecies = list(map(lambda x: x.get(rank,''),response['results']))
        potentialSpecies = list(filter(lambda x: x != '',potentialSpecies))
        if givenRank == "genus": potentialSpecies = list(filter(lambda x: x.split(" ")[0].lower() == name.lower(),potentialSpecies))
        return (name,list(set(potentialSpecies)))
    return (name,[])



from .services import createIndex,\
                      crushIndexInput,\
                      writeApiResultsToIndex,\
                      ensureUniqueInIndex,\
                      get_synonyms

def synonyms(names):
    index = createIndex(names)
    results = get_synonyms([x[0] for x in crushIndexInput(index)])
    writeApiResultsToIndex(index,results)
    return index



from concurrent import futures
import requests

def safeMapResToJson(results):
    res = []
    for r in results:
        try: res.append(r.json())
        except: res.append({})
    return res

def concurrentExecRequests(urls):
    with futures.ThreadPoolExecutor(max_workers=20) as executor:
        res = executor.map(executeAPICall,urls)
    return list(res)

def executeAPICall(url):
    return requests.get(url)


from ..externalAPIs import Gbif
import itertools

def mapKnownTaxaLevelToSpecies(nameAndTaxaTuples):
    try:
        names = list(map(lambda x: x[0], nameAndTaxaTuples))
        taxaTuples = list(map(lambda x: x[1], nameAndTaxaTuples))
        speciesNamesOnly = Gbif.higherTaxaToSpecies(taxaTuples)
        speciesNamesOnly = list(zip(names,speciesNamesOnly))
        return speciesNamesOnly
    except Exception as e:
        print("Caught fatal exception in taxonomic level resolver")
        print(str(e))
        return list(map(lambda x: (x[0],[]), nameAndTaxaTuples))


from .services import scientificToCommon,tagInputWithTaxaLevel,createIndex,mapKnownTaxaLevelToSpecies,crushIndexInput,writeApiResultsToIndex,ensureUniqueInIndex

def to_common(uncheckedNames,includeCommon=True):
    index = createIndex(uncheckedNames)

    results = tagInputWithTaxaLevel([x[0] for x in crushIndexInput(index)])
    writeApiResultsToIndex(index,results)
    results = mapKnownTaxaLevelToSpecies([x for x in crushIndexInput(index) if x[1][0] not in ['species','']])
    results = list(map(lambda x: (x[0],x[1][2]),results))
    writeApiResultsToIndex(index,results)
    for item in index:
        cleanedName, resultSoFar = index[item]
        if len(resultSoFar) == 0:
            index[item] = [cleanedName,[]]
        elif resultSoFar[0] == 'species':
            index[item] = [cleanedName,[resultSoFar[1]]]
        elif resultSoFar == ('',''):
            index[item] = [cleanedName,[]]

    failedDirectScientific = [x[0] for x in crushIndexInput(index) if len(x[1]) > 0]
    results = scientificToCommon(failedDirectScientific)
    writeApiResultsToIndex(index,results)
    
    ensureUniqueInIndex(index)

    return index


from .to_scientific import to_scientific
from .to_species import to_species
from .to_common import to_common
from .classify import classify
from .lowest_rank_intersection import lowest_rank_intersection
from .downstream import downstream
from .synonyms import synonyms
from .children import children
from .upstream import upstream


from .classify import classify
from .services import createIndex,crushIndexInput,writeApiResultsToIndex,ensureUniqueInIndex

def lowest_rank_intersection(species):
    index = createIndex(species)
    for item in index:
        ranks = classify([index[item][0]])
        selectedSpeciesRank = ranks[index[item][0]][1]
        index[item] = [index[item][0],selectedSpeciesRank]
    
    data = list(map(lambda x: set(index[x][1].items()),index.keys()))
    common = list(set.intersection(*data))
    return selectLowestCategory(common)

def selectLowestCategory(intersected):
    rankVals = {}
    intersected = dict(intersected) 
    validRanks = ['species','genus','family','order','class','phylum','kingdom','domain']
    for item in validRanks:
        if item in intersected:
            return (item,intersected[item]) 

    return intersected[-1]



taxa_ordering = ['species','genus','family','order','class','phylum','kingdom','domain']

most_complete_ordering = ['species','genus','tribe','family','order','cohort','legion','division','class','phylum','kingdom','realm','domain']


from .services import commonToScientific,\
                      tagInputWithTaxaLevel,\
                      createIndex,\
                      mapKnownTaxaLevelToSpecies,\
                      crushIndexInput,\
                      writeApiResultsToIndex,\
                      ensureUniqueInIndex,taxa_ordering,\
                      findDownstreamChildren

from .classify import classify
from .downstream import downstream
from collections import defaultdict

def upstream(names,rank):
    index = createIndex(names)
    handlingRank = nextHighestRank(rank)
    classifications = crushIndexInput(classify(names))
    classifications = list(map(lambda x: (x[0],x[1][handlingRank]),classifications))
    writeApiResultsToIndex(index,classifications)

    results = crushIndexInput(index)
    namesOnly = [x[0] for x in results]
    handlingRankInput = [(handlingRank,x[1]) for x in results]
    results = list(zip(namesOnly,[x[1] for x in findDownstreamChildren(handlingRankInput,rank)]))
    writeApiResultsToIndex(index,results)
    return index

def nextHighestRank(rank):
    following = "species"
    newTaxaOrder = [x for x in taxa_ordering]
    newTaxaOrder.append("species")
    for i in range(len(newTaxaOrder)):
        if i == len(newTaxaOrder) - 1: break
        following = newTaxaOrder[i+1]
        if newTaxaOrder[i] == rank:
            break
    return following




import os
import pickle
import csv 
import itertools
from .dataCleaning import cleanSingleSpeciesString
from .constants import taxa_ordering

def writeApiResultsToIndex(index,taxonomyAPIResults):
    taxaAPIDict = {}
    newIndex = {}
    for cleanedName,result in taxonomyAPIResults:
        taxaAPIDict[cleanedName] = result
    
    for name in index:
        newIndex[name] = index[name]
        if index[name][0] in taxaAPIDict:
            newIndex[name][1] = taxaAPIDict[index[name][0]]
    
    return newIndex
    
def ensureUniqueInIndex(index):
    for name in index:
        species = index[name][1]
        species = list(set(species))
        index[name][1] = species

def createIndex(uncheckedNames):
    index = {}
    for item in uncheckedNames:
        index[item] = [cleanSingleSpeciesString(item),'']
    
    return index
        
def crushIndexInput(index):
    return list(map(lambda x: (index[x][0],index[x][1]),index.keys()))


from ..externalAPIs import Itis


def get_synonyms(names):
    try:
        tsns = Itis.retrieveTSNs(names)
        synonyms = Itis.getSynonymsFromTSNs([x[1] for x in tsns])
        tsnToNameMap = dict(list(map(lambda x: (x[1],x[0]),tsns)))
        return list(map(lambda x: (tsnToNameMap[x[0]],x[1]),synonyms))
    except Exception as e:
        print("Caught fatal exception in common name resolver")
        print(str(e))
        return list(map(lambda x: (x,[]),names))



from .services import commonToScientific, \
                      tagInputWithTaxaLevel, \
                      createIndex, \
                      mapKnownTaxaLevelToSpecies, \
                      crushIndexInput, \
                      writeApiResultsToIndex, \
                      ensureUniqueInIndex, \
                      profile_taxonomy, \
                      highest_count_wins_multi_cat, \
                      findDownstreamChildren
from .speciesValidation import sanityCheckOutput
from collections import Counter

def to_species(uncheckedNames,includeCommon=True,sanityCorrect=False):
    index = createIndex(uncheckedNames)
    results = tagInputWithTaxaLevel([x[0] for x in crushIndexInput(index)])
    writeApiResultsToIndex(index,results)
    higherTaxaOnly = [x for x in crushIndexInput(index) if x[1][0] not in ['species','']]
    results = list(findDownstreamChildren(higherTaxaOnly,'species'))
    writeApiResultsToIndex(index,results)
    for item in index:
        cleanedName, resultSoFar = index[item]
        if len(resultSoFar) == 0:
            index[item] = [cleanedName,[]]
        elif resultSoFar[0] == 'species':
            index[item] = [cleanedName,[resultSoFar[1]]]
        elif resultSoFar == ('',''):
            index[item] = [cleanedName,[]]
    failedDirectScientific = [x[0] for x in crushIndexInput(index) if len(x[1]) == 0]
    results = commonToScientific(failedDirectScientific)
    writeApiResultsToIndex(index,results)
    ensureUniqueInIndex(index)
    if sanityCorrect: 
        index = sanityCheckOutput(index)
    return index

def second(nameRankTuples):
    return list(map(lambda x: x[1], nameRankTuples))

def first(nameRankTuples):
    return list(map(lambda x: x[0], nameRankTuples))



from ..externalAPIs import Gbif,Itis
from .constants import most_complete_ordering

def findDownstreamChildren(nameRankTuples,targetRank):
    realNames = list(map(lambda x: x[0],nameRankTuples))
    nameRankTuples = list(map(lambda x: x[1],nameRankTuples))
    remapper = dict(zip(list(map(lambda x: x[1],nameRankTuples)),realNames))
    targetRank = targetRank.lower()
    nameRankTuples = addWorkingBlockToNameRankTuples(nameRankTuples)
    finalRankMapping = prepareFinalRankMapping(nameRankTuples,targetRank)
    failsafe = 10
    while failsafe > 0 and len(nameRankTuples) > 0:
        for k,item in enumerate(nameRankTuples):
            rank,name,workingNames = item
            if len(workingNames) == 0: continue
            tsns = Itis.retrieveTSNs(workingNames)
            tsns = list(filter(lambda x: x[1] != '-1',tsns)) # error code is (name,-1)
            if len(tsns) == 0: 
                nameRankTuples[k] = (rank,name,[])
                continue
            fastCheckWorkingNames = set(list(map(lambda x: x.lower(),workingNames)))
            childrenWithTsns = Itis.downstreamToNextRank([x[1] for x in tsns])
            holding = []
            highestRank = rank
            for childrenResponsesWithTsn in childrenWithTsns:
                tsn, childrenWithRanks = childrenResponsesWithTsn
                childrenWithRanks = list(set(childrenWithRanks))
                childrenWithRanks = list(map(lambda x: (x[0].lower(),x[1]),childrenWithRanks))
                childrenWithRanks = list(filter(lambda x: validReturnedName(x,fastCheckWorkingNames,rank),childrenWithRanks))
                for childRank,childName in childrenWithRanks:
                    highestRank = getHighestChildRank(highestRank,childRank)
                    if childRank == targetRank:
                        finalRankMapping[name].append(childName)
                    else:
                        holding.append(childName)
            nameRankTuples[k] = (highestRank,name,holding)
        
        nameRankTuples = filterCompletedSpecies(nameRankTuples)
        failsafe = failsafe - 1

    preRealNameTranslation = finalRankMapping.items()
    return list(map(lambda x: (remapper[x[0]],x[1]),preRealNameTranslation))

def filterCompletedSpecies(nameRankTuples):
    newNameRankTuples = []
    for rank,name,workingNames in nameRankTuples:
        if len(workingNames) > 0:
            newNameRankTuples.append((rank,name,workingNames))
    return newNameRankTuples

def addWorkingBlockToNameRankTuples(nameRankTuples):
    nameRankTuples = list(set(nameRankTuples))
    nameRankTuples = list(map(lambda x: (x[0].lower(),x[1],[x[1]]),nameRankTuples))
    return nameRankTuples

def prepareFinalRankMapping(nameRankTuples,targetRank):
    finalRankMapping = {}
    for rank,name,workingNames in nameRankTuples:
        finalRankMapping[name] = []
        if rank == targetRank:
            finalRankMapping[name] = workingNames
    return finalRankMapping

def validReturnedName(singleAPITupleOutput,workingNames,rank):
    return singleAPITupleOutput[1].strip().lower() not in workingNames \
        and  transformChildRanks((singleAPITupleOutput[0],''))[0] != '' \
        and  transformChildRanks((rank,''))[0] != '' \
        and most_complete_ordering.index(transformChildRanks((singleAPITupleOutput[0],''))[0]) <= most_complete_ordering.index(transformChildRanks((rank,''))[0])
            
def getHighestChildRank(ORank1,ORank2):
    rank1 = transformChildRanks((ORank1,''))[0]
    rank2 = transformChildRanks((ORank1,''))[0]
    result = most_complete_ordering.index(rank1) < most_complete_ordering.index(rank2)
    if result: return ORank2
    return ORank1
            
def transformChildRanks(nameRankTuple):
    rank,name = nameRankTuple
    for item in most_complete_ordering:
        if item in rank:
            return (item,name)
    return ('',name)



import itertools 
import operator
from .services import createIndex,crushIndexInput,writeApiResultsToIndex,ensureUniqueInIndex,profile_taxonomy

def classify(uncheckedNames):
    index = createIndex(uncheckedNames)
    results = profile_taxonomy([x[0] for x in crushIndexInput(index)])
    results = list(map(runConsensusForSingleSpecies,results))
    writeApiResultsToIndex(index,results)
    return index

def runConsensusForSingleSpecies(nameAndResTuple):
    finalMapping = {}
    name,individualDictionaryMappings = nameAndResTuple
    remove = set(['','no rank'])
    ranks = list(set(itertools.chain(*list(map(lambda x: x.keys(),individualDictionaryMappings)))) - remove)
    for taxaRank in ranks:
        finalMapping[taxaRank] = runConsensusOnSingleTaxa(taxaRank,individualDictionaryMappings)
    return (name,finalMapping)

def runConsensusOnSingleTaxa(taxaRank,individualDictionaryMappings):
    allItemsOfTaxa = []
    for singleMapping in individualDictionaryMappings:
        val = singleMapping.get(taxaRank,'')
        if len(val) > 0: allItemsOfTaxa.append(val)
    
    if len(allItemsOfTaxa) == 0: return ''
    return mostCommonInList(allItemsOfTaxa)

def warnSpeciesNameFailure(individualResult):
    print("Could not index " + str(individualResult['supplied_name_string']))

def mostCommonInList(L):
  SL = sorted((x, i) for i, x in enumerate(L))
  groups = itertools.groupby(SL, key=operator.itemgetter(0))
  def _auxfun(g):
    item, iterable = g
    count = 0
    min_index = len(L)
    for _, where in iterable:
      count += 1
      min_index = min(min_index, where)
    return count, -min_index
  return max(groups, key=_auxfun)[0]


from .services import *
from .tagWithTaxaLevel import tagInputWithTaxaLevel#,mapKnownTaxaLevelToSpecies,mapToCommonName
from .mapHigherTaxaToSpecies import mapKnownTaxaLevelToSpecies
from .mapCommonToScientific import commonToScientific
from .mapScientificToCommon import scientificToCommon
from .profileTaxonomy import profile_taxonomy
from .taxaSynonyms import get_synonyms
from .findDownstream import findDownstreamChildren
from .constants import *
from .consensus import short_circuit,highest_count_wins_multi_cat


import itertools
from .apiUtils import safeMapResToJson, concurrentExecRequests
from .gnrTools import taxaProfiling

class Gnr:
    def __init__(self):
        pass

    # [Names] => [SuccessfulResponses], [FailedNames]
    @staticmethod
    def scientificNameToFullTaxonomy(names):
        capnames = list(map(lambda x: x.capitalize(),names))
        urls = list(map(taxaProfiling.constructUrls,capnames))
        results = concurrentExecRequests(urls)
        results = safeMapResToJson(results)
        results = list(zip(names,results))
        results = list(map(taxaProfiling.processIndividualResult,results))
        results = list(map(taxaProfiling.parseSingleTaxonomyFromAPI,results))
        return results
    




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
     name='EcoNameTranslator',  
     version='2.01',
     author="Daniel Davies",
     author_email="dd16785@bristol.ac.uk",
     description="A lightweight but powerful package for full management and translation of ecological names",
     long_description_content_type='text/markdown',
     long_description=long_description,
     url="https://github.com/Daniel-Davies/MedeinaTranslator",
     packages=setuptools.find_packages(),
     classifiers=[
         "Programming Language :: Python :: 3",
         "License :: OSI Approved :: MIT License",
         "Operating System :: OS Independent",
     ],
     install_requires=['taxon-parser'],
 )


from ..externalAPIs import Gnr
import itertools
from .services import writeApiResultsToIndex
from .consensus import backed_up

def tagInputWithTaxaLevel(names):
    try:
        taxonomyAPIResults = Gnr.scientificNameToFullTaxonomy(names)
        return list(map(backed_up,taxonomyAPIResults))
    except Exception as e:
        print("Caught fatal exception in taxonomic indexer")
        print(str(e))
        return list(map(lambda x: (x,('',x)),names))


    



import itertools 

def constructUrls(name):
    return "https://www.itis.gov/ITISWebService/jsonservice/getITISTermsFromScientificName?srchKey="+name

def processIndividualResult(nameAndResTuple):
    name,res = nameAndResTuple
    if 'itisTerms' not in res: return []
    if res['itisTerms'] is None: return []
    if res['itisTerms'][0] is None: return []
    nameTranslation = res['itisTerms']
    resultsToReturn = []
    for item in nameTranslation:
        if 'commonNames' not in item: continue 
        if item['commonNames'] is None: continue 
        cleanResults = list(filter(lambda x: x is not None,item['commonNames']))
        cleanResults = list(map(lambda x: " ".join(list(map(lambda y: y.capitalize(),x.split(" ")))),cleanResults))
        resultsToReturn.append(cleanResults)
    
    return (name,list(itertools.chain(*resultsToReturn)))




def constructUrls(id_):
    return "https://www.itis.gov/ITISWebService/jsonservice/getSynonymNamesFromTSN?tsn="+id_

def processIndividualResult(idAndResTuple):
    id_,res = idAndResTuple
    errorCode = (id_,[])
    results = []
    if 'synonyms' not in res: return errorCode
    if res['synonyms'] is None: return errorCode
    if res['synonyms'][0] is None: return errorCode
    if len(res['synonyms']) == 0: return errorCode

    for item in res['synonyms']:
        if 'sciName' not in item: continue
        if item['sciName'] is None: continue
        results.append(item['sciName'])
    
    return (id_,results)